"""
Handling of PSG's Global Emission Spectra (GlobES) application
"""

import numpy as np


class GCM:
    """
    Global Circulation Model (GCM)

    Parameters
    ----------
    header : str
        The header of the GCM. This tells PSG how to interpret the data.
    dat : np.ndarray
        The data of the GCM.
    """
    KEY = 'ATMOSPHERE-GCM-PARAMETERS'
    BIN_KEY = 'BINARY'
    ENCODING = 'UTF-8'

    def __init__(
        self,
        header: str,
        dat: np.ndarray
    ):
        self.header = header
        self.dat = dat

    @classmethod
    def from_cfg(cls, d: dict):
        """
        Read a GCM from a config dict.

        Parameters
        ----------
        d : dict
            A dictionary read from a PSG config file.
        """
        if cls.KEY not in d:
            return None
        if cls.BIN_KEY not in d:
            return None
        header = d[cls.KEY]
        bin_dat: bytes = d[cls.BIN_KEY]
        dat = np.frombuffer(bin_dat, dtype=np.float32)
        return cls(header, dat)

    @property
    def content(self) -> bytes:
        """
        Get the content of the GCM in a format that PSG can read.

        Returns
        -------
        bytes
            The content of the GCM.
        """
        params_line = bytes(f'<{self.KEY}>{self.header}\n', encoding=self.ENCODING)
        start_tag = bytes(f'<{self.BIN_KEY}>', encoding=self.ENCODING)
        end_tag = bytes(f'</{self.BIN_KEY}>', encoding=self.ENCODING)
        dat = self.dat.tobytes(order='C')
        return params_line + b'\n' + start_tag + dat + end_tag



"""
Test the `pypsg.docker` module.
"""

import time
import pytest

from pypsg import docker as psgdocker


@pytest.mark.skip(reason='Depends on the test environment')
def test_is_psg_installed():
    """
    Test the `is_psg_installed` function.
    """
    assert psgdocker.is_psg_installed()


@pytest.mark.skip(reason='Depends on the test environment')
def test_is_psg_running():
    """
    Test the `is_psg_running` function.
    """
    assert psgdocker.is_psg_running()


@pytest.mark.skipif(not psgdocker.is_psg_installed(), reason='Depends on the test environment')
def test_psg_start_stop():
    """
    Test the `start_psg` function.
    """
    started_out_running = psgdocker.is_psg_running()
    psgdocker.start_psg()
    assert psgdocker.is_psg_running()
    psgdocker.stop_psg()
    assert not psgdocker.is_psg_running()
    psgdocker.start_psg()
    assert psgdocker.is_psg_running()
    if not started_out_running:
        psgdocker.stop_psg()
    assert started_out_running == psgdocker.is_psg_running()
    # give the container time to setup. This is important for other tests
    time.sleep(1)



"""
Python representation of rad files.
"""

from pathlib import Path
import re
import numpy as np

import astropy.units as u
from astropy import table

from pypsg import settings



class BinRad:
    """
    Binary representation of rad files.
    """
    def __init__(self,content:bytes):
        self.content = content
    @classmethod
    def from_file(cls,path:Path):
        with open(path,'rb') as file:
            content = file.read()
        return cls(content=content)


class PyRad(table.QTable):
    """
    Python representation of rad files.
    """
    SPEC_UNIT = 'SPECUNIT'
    RAD_UNIT = 'RADUNIT'
    _NAMES = 'NAMES'
    @staticmethod
    def _get_metadata(header:str):
        metadata = {}
        
        # metadata['type'] = re.findall(
        #     r'#[ ]+([\w]+ spectrum)',
        #     header
        # )[0]
        # metadata['source'] = re.findall(
        #     r'spectrum\n#(.+)\n',
        #     header
        # )[0]
        # metadata['date'] = re.findall(
        #     r'# Synthesized on (.+)\n',
        #     header
        # )[0]
        
        # vel_unit = u.Unit(
        #     re.findall(r'Doppler velocities \[(.+)\] ',header)[0]
        # )
        # vel_names = re.findall(
        #     r'Doppler velocities \[.+\] \((.+)\)',
        #     header
        # )[0].split(',')
        # vel_values = re.findall(
        #     r'Doppler velocities.+:(.+)\n',
        #     header
        # )[0].split(',')
        # for name,value in zip(vel_names,vel_values):
        #     metadata[name] = float(value) * vel_unit
        
        # metadata['spectral_model'] = re.findall(
        #     r'Spectra synthesized with the (.+)\n',
        #     header
        # )[0]
        # metadata['scattering_method'] = re.findall(
        #     r'Spectra synthesized with the .+\n# (.+)\n',
        #     header
        # )[0]
        
        metadata[PyRad.SPEC_UNIT] = u.Unit(
            re.findall(r'Spectral unit:.+\[(.+)\]\n',header)[0]
        )
        metadata[PyRad.RAD_UNIT] = u.Unit(
            re.findall(r'Radiance unit:.+\[(.+)\]\n',header)[0]
        )
        metadata[PyRad._NAMES] = re.findall(
            r'# (Wave\/freq.+)',
            header
        )[0]
        return metadata
        
    @classmethod
    def from_bytes(cls,b:bytes):
        lines = b.split(b'\n')
        header = [line.decode(settings.get_setting('encoding')) for line in lines if line.startswith(b'#')]
        content = [line.decode(settings.get_setting('encoding')) for line in lines if (not line.startswith(b'#') and len(line)>0)]
        
        metadata = cls._get_metadata('\n'.join(header))
        
        spectral_unit:u.Unit = metadata[cls.SPEC_UNIT]
        radiance_unit:u.Unit = metadata[cls.RAD_UNIT]
        
        names = metadata[cls._NAMES].split(' ')
        content = np.array(
            [
                np.fromstring(line,sep=' ') for line in content
            ]
        )
        data = {}
        for i, name in enumerate(names):
            data[name] = content[:,i] * (spectral_unit if i==0 else radiance_unit)
        return cls(data=data)
    @property
    def wl(self):
        return self['Wave/freq']


"""
PyPSG Requests
--------------

Direct access to the PSG API
"""
from typing import Union, Dict
import re
import requests

from pypsg.cfg import PyConfig, BinConfig
from pypsg import settings
from pypsg.rad import PyRad
from pypsg.lyr import PyLyr

typedict: Dict[bytes, Union[PyConfig, PyRad, PyLyr]] = {
    b'cfg': PyConfig,
    b'rad': PyRad,
    b'lyr': PyLyr,
    b'noi': PyRad
}


class PSGResponse:
    """
    A class to parse the response from the PSG API.
    
    Parameters
    ----------
    cfg : PyConfig
        The PSG .cfg file.
    rad : PyRad
        The PSG .rad file.
    lyr : PyLyr
        The PSG .lyr file.
    noi : PyRad
        The PSG .noi file.
    """
    def __init__(
        self,
        cfg: PyConfig = None,
        rad: PyRad = None,
        lyr: PyLyr = None,
        noi: PyRad = None
    ):
        self.cfg = cfg
        self.rad = rad
        self.lyr = lyr
        self.noi = noi

    @classmethod
    def from_bytes(cls, b: bytes):
        """
        Read the response from PSG as a byte string.
        
        Parameters
        ----------
        b : bytes
            The response from the PSG. This is the returned file read as bytes.
        """
        pattern = rb'results_([\w]+).txt'
        split_text = re.split(pattern, b)
        names = split_text[1::2]
        content = split_text[2::2]
        data = {}
        for name, dat in zip(names, content):
            data[name] = dat.strip()
        kwargs = {}
        for key, value in typedict.items():
            value: PyConfig | PyRad | PyLyr
            if key in data:
                kwargs[key.decode(settings.get_setting('encoding'))] = value.from_bytes(data[key])
        return cls(**kwargs)
    @classmethod
    def null(cls):
        return cls()


class APICall:
    """
    A class to call the PSG API.

    Parameters
    ----------
    cfg : Config
        The PSG configuration.
    output_type : str or None
        The type of output to ask for.
    app : str or None
        The app to use.
    url : str
        The URL to send the request to.

    Attributes
    ----------
    cfg : Config
        The PSG configuration.
    output_type : str or None
        The type of output to ask for.
    app : str or None
        The app to use.
    url : str
        The URL to send the request to.
    """

    def __init__(
        self,
        cfg: Union[BinConfig, PyConfig],
        output_type: str = None,
        app: str = None,
        url: str = None
    ):
        self.cfg = cfg
        self._type = output_type
        self.app = app
        self.url = url
        if self.url is None:
            self.url = settings.get_setting('url')
        self._validate()

    def _validate(self):
        """
        Validate a class instance.

        Raises
        ------
        TypeError
            If self.cfg is not a Config object.
        TypeError
            If self.type is not a string or None.
        TypeError
            If self.app is not a string or None.
        TypeError
            If self.url is not a string.
        """
        if not isinstance(self.cfg, (PyConfig, BinConfig)):
            raise TypeError(
                'apiCall.cfg must be a PyConfig or BinaryConfig object')
        if not (isinstance(self._type, str) or self._type is None):
            if not isinstance(self._type, (list, tuple)):
                raise TypeError('apiCall.type must be a string or None')
            else:
                raise NotImplementedError('Multiple types not implemented. If you know how to do this please open an issue.')
                # for t in self._type:
                #     if not isinstance(t, str):
                #         raise TypeError('apiCall.type must be a string or None')
        if not (isinstance(self.app, str) or self.app is None):
            raise TypeError('apiCall.app must be a string or None')
        if not isinstance(self.url, str):
            raise TypeError('apiCall.url must be a string')

    @property
    def is_single_file(self):
        """
        True if only a single file is expected back from the PSG API.
        
        Returns
        -------
        bool
            True if only a single file is expected back from the PSG API.
        """
        if isinstance(self._type, (tuple, list)):
            return False
        if self._type == 'all':
            return False
        return True
    @property
    def type(self):
        """
        The type of output to ask for.

        :type: str
        """
        if isinstance(self._type,str):
            return self._type
        else:
            return ','.join(self._type)

    def __call__(self) -> PSGResponse:
        """
        Call the PSG API

        Returns
        -------
        bytes
            The reply from PSG.
        """
        data = dict(file=self.cfg.content)
        if self.type is not None:
            data['type'] = self.type
        if self.app is not None:
            data['app'] = self.app
        api_key = settings.get_setting('api_key')
        if api_key is not None:
            data['key'] = api_key
        url = self.url
        if '/api.php' not in url:
            url = f'{url}/api.php'
        reply = requests.post(
            url=url,
            data=data,
            timeout=settings.REQUEST_TIMEOUT
        )
        if reply.status_code != 200:
            raise requests.exceptions.HTTPError(reply.content)
        if self._type in ['upd', 'set']:
            return PSGResponse.null()
        elif not self.is_single_file:
            return PSGResponse.from_bytes(reply.content)
        else:
            returntype = typedict[self._type.encode(settings.get_setting('encoding'))]
            return PSGResponse(**{self._type:returntype.from_bytes(reply.content)})



"""
A module to store PSG config models.
"""
from astropy import units as u
from astropy.units import cds
from astropy.units import imperial

from pypsg import units as u_psg
from pypsg.cfg.base import Model
from pypsg.cfg.base import CharField, IntegerField, DateField
from pypsg.cfg.base import FloatField, QuantityField, CodedQuantityField, CharChoicesField
from pypsg.cfg.base import GeometryOffsetField, MoleculesField, AerosolsField
from pypsg.cfg.base import ProfileField, BooleanField, UnitChoicesField


class Target(Model):
    """
    PSG parameter model for keywords begining with `OBJECT`.
    """
    object = CharChoicesField('object', ('Exoplanet', 'Planet',
                              'Asteroid', 'Moon', 'Comet', 'Object'), max_length=50)
    name = CharField('object-name', max_length=50)
    date = DateField('object-date')
    diameter = QuantityField('object-diameter', u.km)
    gravity = CodedQuantityField(
        allowed_units=(u.Unit('m s-2'), u.Unit('g cm-3'), u.kg),
        unit_codes=('g', 'rho', 'kg'),
        fmt=('.4f', '.4f', '.4e'),
        names=('object-gravity', 'object-gravity-unit')
    )
    star_distance = QuantityField('object-star-distance', u.AU)
    star_velocity = QuantityField('object-star-velocity', u.Unit('km s-1'))
    solar_longitude = QuantityField('object-solar-longitude', u.deg)
    solar_latitude = QuantityField('object-solar-latitude', u.deg)
    season = QuantityField('object-season', u.deg)
    inclination = QuantityField('object-inclination', u.deg)
    position_angle = QuantityField('object-position-angle', u.deg)
    star_type = CharChoicesField(
        'object-star-type', ('O', 'B', 'A', 'F', 'G', 'K', 'M', '', '-'), max_length=1)
    star_temperature = QuantityField('object-star-temperature', u.K)
    star_radius = QuantityField('object-star-radius', u.R_sun)
    star_metallicity = FloatField('object-star-metallicity')
    obs_longitude = QuantityField('object-obs-longitude', u.deg)
    obs_latitude = QuantityField('object-obs-latitude', u.deg)
    obs_velocity = QuantityField('object-obs-velocity', u.Unit('km s-1'))
    period = QuantityField('object-period', u.day)
    orbit = CharField('object-orbit', max_length=100)


class Geometry(Model):
    """
    PSG parameter model for fields starting with `GEOMETRY`.
    """
    geometry = CharField('geometry', max_length=20)
    ref = CharField('geometry-ref', max_length=50)
    offset = GeometryOffsetField()
    observer_altitude = CodedQuantityField(
        allowed_units=(u.AU, u.km, u_psg.diameter, u.pc),
        unit_codes=('AU', 'km', 'diameter', 'pc'),
        fmt='.4f',
        names=('geometry-obs-altitude', 'geometry-altitude-unit')
    )
    azimuth = QuantityField('geometry-azimuth', u.deg)
    stellar_type = CharChoicesField(
        'geometry-stellar-type', ('O', 'B', 'A', 'F', 'G', 'K', 'M', ''), max_length=1)
    stellar_temperature = QuantityField('geometry-stellar-temperature', u.K)
    stellar_magnitude = FloatField('geometry-stellar-magnitude')
    # GEOMETRY-OBS-ANGLE -- Computed by PSG
    # GEOMETRY-SOLAR-ANGLE -- Computed by PSG
    disk_angles = IntegerField('geometry-disk-angles')
    # GEOMETRY-PLANET-FRACTION -- Computed by PSG
    # GEOMETRY-STAR-FRACTION -- Computed by PSG
    # GEOMETRY-STAR-DISTANCE -- Computed by PSG
    # GEOMETRY-ROTATION -- Computed by PSG
    # GEOMETRY-BRDFSCALER -- Computed by PSG

    def _type_to_create(self, *args, **kwargs):
        cfg = kwargs.get('cfg')
        geometry = self.geometry.read(cfg)
        if geometry is None:
            return Geometry
        if geometry == 'Observatory':
            return Observatory
        elif geometry == 'Nadir':
            return Nadir
        elif geometry == 'Limb':
            return Limb
        else:
            raise ValueError(f'Unknown geometry type {geometry}')


class Observatory(Geometry):
    """
    Observatory Geometry
    """

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.geometry.value = 'Observatory'


class Nadir(Geometry):
    """
    Nadir Geometry
    """
    zenith = QuantityField('geometry-user-param', u.deg)

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.geometry.value = 'Nadir'


class Limb(Geometry):
    """
    Limb Geometry
    """
    limb_altitude = QuantityField('geometry-user-param', u.km)

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.geometry.value = 'Limb'


class Occultation(Geometry):
    """
    Occultation Geometry
    """
    occultation_altitude = QuantityField('geometry-user-param', u.km)

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.geometry.value = ''
        raise NotImplementedError(
            'I don\'t know the keyword for this geometry')


class LookingUp(Geometry):
    """
    Looking Up Geometry
    """

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.geometry.value = ''
        raise NotImplementedError(
            'I don\'t know the keyword for this geometry')


class Atmosphere(Model):
    """
    Base Atmosphere model
    """
    structure = CharChoicesField(
        'atmosphere-structure', ('None', 'Equilibrium', 'Coma'))

    def _type_to_create(self, *args, **kwargs):
        cfg = kwargs.get('cfg')
        structure = self.structure.read(cfg)
        if structure is None:
            return Atmosphere
        if structure == 'None':
            return NoAtmosphere
        elif structure == 'Equilibrium':
            return EquilibriumAtmosphere
        elif structure == 'Coma':
            return ComaAtmosphere
        else:
            raise ValueError(f'Unknown atmosphere type {structure}')


class NoAtmosphere(Atmosphere):
    """
    No Atmosphere
    """
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.structure.value = 'None'


class EquilibriumAtmosphere(Atmosphere):
    """
    An atmosphere that is in hydrostatic equilibrium.
    """
    pressure = CodedQuantityField(
        # pylint: disable-next=no-member
        allowed_units=(u.Pa, u.bar, u_psg.kbar, u_psg.mbar,
                       # pylint: disable-next=no-member
                       u_psg.ubar, cds.atm, u.torr, imperial.psi),
        unit_codes=('Pa', 'bar', 'kbar', 'mbar', 'ubar',
                    'atm', 'torr', 'psi'),  # what is `at`?
        fmt='.4e', names=('atmosphere-pressure', 'atmosphere-punit')
    )
    temperature = QuantityField('atmosphere-temperature', u.K)
    weight = QuantityField('atmosphere-weight', u.Unit('g mol-1'))
    continuum = CharField('atmosphere-continuum', max_length=300)
    molecules = MoleculesField()
    aerosols = AerosolsField()
    nmax = IntegerField('atmosphere-nmax')
    lmax = IntegerField('atmosphere-lmax')
    description = CharField('atmosphere-description', max_length=200)
    profile = ProfileField()

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.structure.value = 'Equilibrium'


class ComaAtmosphere(Atmosphere):
    """
    An atmosphere that is the result of outgassing.
    """
    gas_production = QuantityField('atmosphere-pressure', u.Unit('s-1'))
    at_1au = BooleanField('atmosphere-punit', true='gasau', false='gas')
    expansion_velocity = QuantityField('atmosphere-weight', u.Unit('m s-1'))
    continuum = CharField('atmosphere-continuum', max_length=300)
    molecules = MoleculesField()
    aerosols = AerosolsField()
    nmax = IntegerField('atmosphere-nmax')
    lmax = IntegerField('atmosphere-lmax')
    tau = QuantityField('atmosphere-tau', u.s)
    description = CharField('atmosphere-description', max_length=200)
    profile = ProfileField()

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.structure.value = 'Coma'


class Surface(Model):
    """
    The surface model.
    """
    temperature = QuantityField('surface-temperature', u.K)
    albedo = FloatField('surface-albedo')
    emissivity = FloatField('surface-emissivity')


class Generator(Model):
    """
    Combines the source and instrument models into spectra.
    """
    resolution_kernel = BooleanField('generator-resolutionkernel')
    gas_model = BooleanField('generator-gas-model')
    continuum_model = BooleanField('generator-cont-model')
    continuum_stellar = BooleanField('generator-cont-stellar')
    apply_telluric_noise = BooleanField('generator-trans-show')
    apply_telluir_obs = BooleanField('generator-trans-apply')
    telluric_params = CharField('generator-trans', max_length=20)
    rad_units = UnitChoicesField(
        'generator-radunits',
        options=tuple(value for value in u_psg.radiance_units.values()),
        codes = tuple(key for key in u_psg.radiance_units)
    )
    log_rad = BooleanField('generator-lograd')
    gcm_binning = IntegerField('generator-gcm-binning')


class Telescope(Model):
    """
    Base class for telescopes.
    """
    telescope = CharChoicesField(
        'generator-telescope',
        options=('SINGLE', 'ARRAY', 'CORONA', 'AOTF', 'LIDAR')
    )
    apperture = QuantityField('generator-diamtele', u.Unit('m'))
    zodi = FloatField('generator-telescope2')
    fov = CodedQuantityField(
        allowed_units=(u.arcsec, u.arcmin, u.deg, u.km,
                       u_psg.diameter, u_psg.diffraction),
        unit_codes=('arcsec', 'arcmin', 'deg', 'km', 'diameter', 'diffrac'),
        fmt='.4e',
        names=('generator-beam', 'generator-beamunit')
    )
    range1 = CodedQuantityField(
        allowed_units=(u.um, u.nm, u.mm, u.AA, u.Unit('cm-1'),
                       u.MHz, u.GHz, u.kHz),
        unit_codes=('um', 'nm', 'mm', 'An', 'cm', 'MHz', 'GHz', 'kHz'),
        fmt='.4e', names=('generator-range1', 'generator-rangeunit')
    )
    range2 = CodedQuantityField(
        allowed_units=(u.um, u.nm, u.mm, u.AA, u.Unit('cm-1'),
                       u.MHz, u.GHz, u.kHz),
        unit_codes=('um', 'nm', 'mm', 'An', 'cm', 'MHz', 'GHz', 'kHz'),
        fmt='.4e', names=('generator-range2', 'generator-rangeunit')
    )
    resolution = CodedQuantityField(
        allowed_units=(u_psg.resolving_power, u.um, u.nm, u.mm, u.AA,
                       u.Unit('cm-1'), u.MHz, u.GHz, u.kHz),
        unit_codes=('RP', 'um', 'nm', 'mm', 'An', 'cm', 'MHz', 'GHz', 'kHz'),
        fmt='.4e', names=('generator-resolution', 'generator-resolutionunit')
    )

    def _type_to_create(self, *args, **kwargs):
        cfg = kwargs['cfg']
        value = self.telescope.read(cfg)
        if value is None:
            return Telescope
        if value == 'SINGLE':
            return SingleTelescope
        elif value == 'ARRAY':
            return Interferometer
        elif value == 'CORONA':
            return Coronagraph
        elif value == 'AOTF':
            return AOTF
        elif value == 'LIDAR':
            return LIDAR
        else:
            raise ValueError(f'Unknown telescope type: {value}')


class Noise(Model):
    """
    Base class for noise models.
    """
    noise_type = CharChoicesField(
        'generator-noise',
        options=('NO', 'TRX', 'RMS', 'BKG', 'NEP', 'D*', 'CCD')
    )
    exp_time = QuantityField('generator-noisetime', u.s)
    n_frames = IntegerField('generator-noiseframes')
    n_pixels = FloatField(
        'generator-noisepixels',
        allow_table=True,
        xunit=u.um,
        yunit=None
    )
    desc = CharField('generator-instrument', max_length=500)

    def _type_to_create(self, *args, **kwargs):
        cfg = kwargs['cfg']
        value = self.noise_type.read(cfg)
        if value is None:
            return Noise
        if value == 'NO':
            return Noiseless
        elif value == 'TRX':
            return RecieverTemperatureNoise
        elif value == 'RMS':
            return ConstantNoise
        elif value == 'BKG':
            return ConstantNoiseWithBackground
        elif value == 'NEP':
            return PowerEquivalentNoise
        elif value == 'D*':
            return Detectability
        elif value == 'CCD':
            return CCD
        else:
            raise ValueError(f'Unknown noise type: {value}')


class SingleTelescope(Telescope):
    """
    A simple telescope with a single apperture.
    
    Handbook
    --------
    This mode is the classical observatory / instrument
    optical setup, in which the etendue is defined by the
    effective collecting area of the main mirror :math:`A_{Tele}`
    and its corresponding solid angle :math:`\\Omega`.
    """
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.telescope.value = 'SINGLE'


class Interferometer(Telescope):
    """
    An interferometry array.
    """
    n_telescopes = IntegerField('generator-telescope1')

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.telescope.value = 'ARRAY'


class Coronagraph(Telescope):
    """
    A coronagraph.
    """
    contrast = FloatField('generator-telescope1')
    iwa = FloatField(
        'generator-telescope3',
        allow_table=True,
        xunit=None,
        yunit=None
    )

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.telescope.value = 'CORONA'


class AOTF(Telescope):
    """
    Acousto-Optical-Tunable-Filter (AOTF)
    """
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.telescope.value = 'AOTF'


class LIDAR(Telescope):
    """
    A laser source is injected into the FOV.
    """
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.telescope.value = 'LIDAR'


class Noiseless(Noise):
    """
    No noise. This is not the same as a `null` value.
    """
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.noise_type.value = 'NO'


class RecieverTemperatureNoise(Noise):
    """
    Receiver temperature (radio).
    """
    temperature = QuantityField('generator-noise1', u.K)
    g_factor = FloatField('generator-noise2')

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.noise_type.value = 'TRX'


class ConstantNoise(Noise):
    """
    Constant noise model.
    """
    sigma = FloatField('generator-noise1')

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.noise_type.value = 'RMS'


class ConstantNoiseWithBackground(Noise):
    """
    No additional description in handbook.
    """
    sigma = FloatField('generator-noise1')

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.noise_type.value = 'BKG'


class PowerEquivalentNoise(Noise):
    """
    Noise Equivalent Power
    """
    sensitivity = QuantityField('generator-noise1', u.W/u.Hz**(1/2))

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.noise_type.value = 'NEP'


class Detectability(Noise):
    """
    Detectivity
    """
    sensitivity = QuantityField('generator-noise1', u.cm*u.Hz**(1/2)/u.W)
    pixel_size = QuantityField('generator-noise2', u.um)

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.noise_type.value = 'D*'


class CCD(Noise):
    """
    Charge image sensor (e.g., CCD, CMOS, EMCCD, ICCD / MCP)
    """
    read_noise = QuantityField(
        'generator-noise1',
        u.electron,
        allow_table=True,
        xunit=u.um,
        yunit=u.electron
    )
    dark_current = QuantityField(
        'generator-noise2',
        u.electron/u.s,
        allow_table=True,
        xunit=u.um,
        yunit=u.electron/u.s
    )
    thoughput = FloatField(
        'generator-noiseoeff',
        allow_table=True,
        xunit=u.um,
        yunit=None
    )
    emissivity = FloatField('generator-noiseoemis')
    temperature = QuantityField('generator-noiseotemp', u.K)
    pixel_depth = QuantityField('generator-noisewell', u.electron)

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.noise_type.value = 'CCD'



"""
PyPSG settings
==============

This module allows users to configure PyPSG.
"""
from pathlib import Path
import warnings
import json

REQUEST_TIMEOUT = 120

PSG_URL = 'https://psg.gsfc.nasa.gov/api.php'
PSG_PORT = 3000
INTERNAL_PSG_URL = f'http://localhost:{PSG_PORT}/api.php'

USER_DATA_PATH = Path.home() / '.pypsg'
USER_SETTINGS_PATH = USER_DATA_PATH / 'settings.json'

DEFAULT_SETTINGS = {
    'url': PSG_URL,
    'api_key': None,
    'encoding': 'utf-8',
    'cfg_max_lines': 1500
}


settings_need_reload = False
def save_settings(**kwargs):
    if not USER_DATA_PATH.exists():
        USER_DATA_PATH.mkdir()
    if not USER_SETTINGS_PATH.exists():
        USER_SETTINGS_PATH.touch()
    with USER_SETTINGS_PATH.open('r') as file:
        try:
            previous_settings = json.load(file)
        except json.decoder.JSONDecodeError:
            previous_settings = {}
        
    for key, value in kwargs.items():
        if key in DEFAULT_SETTINGS.keys():
            previous_settings[key] = value
        else:
            raise KeyError(f'Unknown setting {key}.')
    with USER_SETTINGS_PATH.open('w') as file:
        json.dump(previous_settings, file, indent=4)
    print(f'Saved settings to {USER_SETTINGS_PATH}')
    # pylint: disable-next=global-statement
    global settings_need_reload
    settings_need_reload = True
    print('Reloading settings...')
    reload_settings()

def load_settings():
    try:
        with USER_SETTINGS_PATH.open('r') as file:
            try:
                settings = json.load(file)
            except json.decoder.JSONDecodeError:
                settings = {}
    except FileNotFoundError:
        settings = {}
    for key, value in DEFAULT_SETTINGS.items():
        if key not in settings:
            settings[key] = value
    # pylint: disable-next=global-statement
    global settings_need_reload
    settings_need_reload = False
    return settings

user_settings = load_settings()

def reload_settings():
    # pylint: disable-next=global-statement
    global user_settings
    user_settings = load_settings()

class StaleSettingsWarning(RuntimeWarning):
    pass

def get_setting(key):
    if settings_need_reload:
        msg = 'Your user settings have changed recently.\n'
        msg += 'Please reload the settings using the `pypsg.settings.reload_settings()` function.'
        warnings.warn(msg,StaleSettingsWarning) 
    if key in user_settings:
        return user_settings[key]
    else:
        raise KeyError(f'Unknown setting {key}.')


"""
Helpers for running psg locally
"""
import docker

from . import settings

PSG_LATEST = 'psg:latest'
PSG_CONTAINER_NAME = 'psg'

def is_psg_installed()->bool:
    """
    Determine if a local version of PSG is installed.
    
    This function checks all docker containers and compares their
    `RepoTags` attribute to the string `psg:latest`.
    
    Returns
    -------
    bool
        True if a local version of PSG is installed.
    
    Notes
    -----
    This function also checks the name of the image. This is important
    because the name is used to start/stop the container. The expected
    name is `psg`.
    
    Warning
    -------
    Prior to `v0.1.2`, this function would throw an error if the
    docker engine was not installed. Since `v0.1.2`, this function
    returns `False` if the docker engine is not installed.
    """
    try:
        installed_containers = docker.from_env().containers.list(all=True)
        for container in installed_containers:
            image = container.image
            if PSG_LATEST in image.attrs['RepoTags']:
                if container.name == PSG_CONTAINER_NAME:
                    return True
        return False
    except docker.errors.DockerException:
        return False

def get_psg_container()->docker.models.containers.Container:
    """
    Get the psg container.
    
    Returns
    -------
    docker.models.containers.Container
        The psg container.
    """
    return docker.from_env().containers.get(PSG_CONTAINER_NAME)

class PSGNotInstalledError(Exception):
    """
    Exception raised when a local version of PSG is not installed.
    """

def is_psg_running()->bool:
    """
    Determine if a local version of PSG is running.
    
    Returns
    -------
    bool
        True if a local version of PSG is running.
    """
    try:
        container = get_psg_container()
    except docker.errors.NotFound as e:
        msg = 'PSG is not installed. '
        url = 'https://psg.gsfc.nasa.gov/helpapi.php#installation'
        msg += f'Visit {url} for installation instructions.'
        raise PSGNotInstalledError(msg) from e
    return container.status == 'running'

def start_psg(strict=True):
    """
    Start the psg container.
    
    Parameters
    ----------
    strict : bool, optional
        If True, raise an error if PSG is not installed locally. By default True.
    """
    if not is_psg_installed():
        if strict:
            msg = 'PSG is not installed. '
            url = 'https://psg.gsfc.nasa.gov/helpapi.php#installation'
            msg += f'Visit {url} for installation instructions.'
            raise PSGNotInstalledError(msg)
        else:
            return None
    if not is_psg_running():
        get_psg_container().start()

def stop_psg(strict=True):
    """
    Stop the psg container.
    
    Parameters
    ----------
    strict : bool, optional
        If True, raise an error if PSG is not installed locally. By default True.
    """
    if not is_psg_installed():
        if strict:
            msg = 'PSG is not installed. '
            url = 'https://psg.gsfc.nasa.gov/helpapi.php#installation'
            msg += f'Visit {url} for installation instructions.'
            raise PSGNotInstalledError(msg)
        else:
            return None
    if is_psg_running():
        get_psg_container().stop()

def set_psg_url(internal=True):
    """
    Set the psg URL. This checks if PSG is installed locally first.
    
    Parameters
    ----------
    internal : bool, optional
        If True, use the internal psg URL. By default True.
    """
    if internal and is_psg_installed():
        url = settings.INTERNAL_PSG_URL
    else:
        url = settings.PSG_URL
    settings.save_settings(url=url)

def set_url_and_run():
    """
    Set the psg URL and run the container.
    Again, this checks if PSG is installed locally first.
    
    Parameters
    ----------
    url : str
        The URL to use.
    """
    if is_psg_installed():
        url = settings.INTERNAL_PSG_URL
        start_psg(strict=True)
    else:
        url = settings.PSG_URL
    settings.save_settings(url=url)



"""
PyPSG top-level module
======================
"""
__version__ = '0.1.2'
from pypsg.request import APICall
from pypsg import cfg
from pypsg.cfg import PyConfig
from pypsg.rad import PyRad
from pypsg.lyr import PyLyr
from pypsg import settings
from pypsg import units
from . import docker



"""
Methods to parse config files.
"""
from typing import Union, Type
from pathlib import Path

import warnings

from pypsg.cfg import models, globes
from pypsg import settings


class ConfigTooLongWarning(UserWarning):
    """
    The PSG configuration is too long,
    and may stop updating soon.
    """


class BinConfig:
    """
    PSG configuration structure.
    
    Parameters
    ----------
    content : bytes
        The content of the configuration.
    
    Attributes
    ----------
    enconding : str
        The encoding of the config. Set to 'UTF-8'.
    content : bytes
        The content of the config.
    has_binary : bool
        True if there is a `<BINARY>` section in the config.
    """
    encoding='UTF-8'
    def __init__(self,content:bytes):
        self.content = content
    @classmethod
    def from_file(cls,path:Path):
        """
        Read a config from a file.

        Parameters
        ----------
        path : pathlib.Path
            The path to the file.

        Returns
        -------
        Config
            A config constructed using the provided file.
        """
        # warnings.warn('This method has not been tested.',RuntimeWarning)
        with open(path,'rb') as file:
            content = file.read()
        return cls(content=content)
    @property
    def has_binary(self)->bool:
        """
        True if the config contains a binary section
        
        :type: bool
        """
        # warnings.warn('This method has not been tested.',RuntimeWarning)
        return b'<BINARY>' in self.content
    @property
    def binary(self)->bytes:
        """
        The binary section of the config.
        
        :type: bytes
        """
        # warnings.warn('This method has not been tested.',RuntimeWarning)
        if not self.has_binary:
            raise ValueError('This config contains no binary section.')
        return self.content.split(b'<BINARY>')[1].split(b'</BINARY>')[0]
    @property
    def dict(self)->dict:
        """
        A dictionary with all the keyword, value pairs.

        :type: dict
        """
        # warnings.warn('This method has not been tested.',RuntimeWarning)
        content = self.content
        binary = None
        if self.has_binary:
            content = content.split(b'<BINARY>')[0] + content.split(b'</BINARY>')[1]
            binary = self.binary
        content = str(content,encoding=self.encoding)
        n_lines = len(content.split('\n'))
        if n_lines > settings.get_setting('cfg_max_lines'):
            warnings.warn('The config is too long.',ConfigTooLongWarning)
        cfg = {}
        for line in content.split('\n'):
            if not (line.isspace() or len(line)==0):
                end_of_kwd = line.index('>')+1
                kwd = line[:end_of_kwd].replace('<','').replace('>','')
                val = line[end_of_kwd:]
                cfg[kwd] = val
        if binary is not None:
            cfg['BINARY'] = binary
        return cfg

class PyConfig:
    """
    A configuration in the form of a python object.
    """
    def __init__(
        self,
        target:models.Target = None,
        geometry:models.Geometry = None,
        atmosphere:models.Atmosphere = None,
        surface:models.Surface = None,
        generator:models.Generator = None,
        telescope:models.Telescope = None,
        noise:models.Noise = None,
        gcm:globes.GCM = None
    ):
        self.target:models.Target = target
        if self.target is None:
            self.target = models.Target()
        
        self.geometry:models.Geometry = geometry
        if self.geometry is None:
            self.geometry = models.Geometry()
        
        self.atmosphere:Union[
            models.NoAtmosphere,models.EquilibriumAtmosphere,models.ComaAtmosphere
            ] = atmosphere
        if self.atmosphere is None:
            self.atmosphere = models.Atmosphere()
        
        self.surface:models.Surface = surface
        if self.surface is None:
            self.surface = models.Surface()
        
        self.generator:models.Generator = generator
        if self.generator is None:
            self.generator = models.Generator()
        
        self.telescope:Union[
            models.SingleTelescope,models.Interferometer,models.Coronagraph,
            models.AOTF,models.LIDAR
            ] = telescope
        if self.telescope is None:
            self.telescope = models.Telescope()
        
        self.noise:Union[
            models.Noiseless,models.RecieverTemperatureNoise,
            models.ConstantNoise,models.ConstantNoiseWithBackground,
            models.PowerEquivalentNoise,models.Detectability,models.CCD
            ] = noise
        if self.noise is None:
            self.noise = models.Noise()
        
        self.gcm:globes.GCM | None = gcm
        
    @classmethod
    def from_dict(cls,d:dict):
        return cls(
            target=models.Target.from_cfg(d),
            geometry=models.Geometry.from_cfg(d),
            atmosphere=models.Atmosphere.from_cfg(d),
            surface=models.Surface.from_cfg(d),
            generator=models.Generator.from_cfg(d),
            telescope=models.Telescope.from_cfg(d),
            noise=models.Noise.from_cfg(d),
            gcm=globes.GCM.from_cfg(d)
        )
    @classmethod
    def from_binaryconfig(cls,config:BinConfig):
        return cls.from_dict(config.dict)
    @classmethod
    def from_bytes(cls,config:bytes):
        return cls.from_binaryconfig(BinConfig(config))
    @classmethod
    def from_file(cls,path:Path):
        return cls.from_binaryconfig(BinConfig.from_file(path))
    @property
    def content(self)->bytes:
        lines = []
        for model in [
            self.target,
            self.geometry,
            self.atmosphere,
            self.surface,
            self.generator,
            self.telescope,
            self.noise
        ]:
            c = model.content
            if c != b'':
                lines.append(c)
        if self.gcm is not None:
            lines.append(self.gcm.content)
        return b'\n'.join(lines)
    def to_file(self,path:Path):
        """
        Write the config to a file.
        
        Parameters
        ----------
        path : pathlib.Path
            The path to the file.
        """
        with open(path,'wb') as f:
            f.write(self.content)


"""
PSG Layer files
"""
from typing import Tuple, List
from pathlib import Path
import re
import numpy as np
from astropy import units as u
from astropy.table import QTable
from astropy.io import fits

from pypsg.cfg.base import Profile
from pypsg import settings

class PyLyr:
    """
    Python representation of lyr files.
    """
    PROF_EXT = 'PROFILE'
    CG_EXT = "CG"
    def __init__(
        self,
        prof:QTable,
        cg:QTable
    ):
        self.prof = prof
        self.prof.meta['EXTNAME'] = self.PROF_EXT
        self.cg = cg
        self.cg.meta['EXTNAME'] = self.CG_EXT
    
    @staticmethod
    def _parse(text:str)->Tuple[dict,str,str,str]:
        metadata = {}
        other_data = {}
        def find(s:str):
            """TODO: Change all below to use this func.
            If something break this is probably why."""
            try:
                return re.findall(s,text)[0]
            except IndexError:
                return None
        # metadata['source'] = re.findall(r'# (NASA-GSFC.*)',text)[0]
        # metadata['scattering'] = re.findall(r'# (.*), LMAX',text)[0]
        # metadata['date'] = re.findall(r'# Synthesized on (.*)',text)[0]
        # metadata['rad_tran_method'] = re.findall(r'# Radiative transfer method: (.*)',text)[0]
        # syn_res_unit = u.Unit(
        #     re.findall(r'# Synthesis resolution \[(.*)\]',text)[0]
        # )
        # syn_res_val, l_over_dl, points, fine_scalar =  re.findall(
        #     r'# Synthesis resolution .*: (.*)',text
        # )[0].split(' ')
        # metadata['syn_res'] = float(syn_res_val) * syn_res_unit
        # metadata['l_over_dl'] = float(l_over_dl)
        # metadata['points'] = float(points)
        # metadata['fine_scalar'] = float(fine_scalar)
        
        # metadata['profile_source'] = re.findall(r'# Molecular abundance profile: (.*)',text)[0]
        metadata['molecules'] = re.findall(r'# Molecules considered: (.*)',text)[0]
        metadata['molec_sources'] = re.findall(r'# Molecular sources: (.*)',text)[0]
        metadata['molec_abuns'] = re.findall(r'# Molecular abundances: (.*)',text)[0]
        metadata['molec_abun_units'] = re.findall(r'# Molecular abundance units: (.*)',text)[0]
        # metadata['collisional'] = re.findall(r'# Collissional partners (.*)',text)[0]
        # metadata['uv_cross_sections'] = find(r'UV cross-sections included: (.*)')
        # metadata['cia'] = re.findall(r'# Collision-Induced-Absorption: (.*)',text)[0]
        # metadata['molec_rayleigh'] = re.findall(r'# Molecular Rayleigh (.*)',text)[0]
        # rmax_minus_1, bending = re.findall(r'# Refraction .*: (.*)',text)[0].split(' ')
        # bending_unit = u.Unit(re.findall(r'# Refraction .* bending \[(.*)\]',text)[0])
        # metadata['rmax_minus_1'] = float(rmax_minus_1)
        # metadata['bending'] = float(bending) * bending_unit
        
        metadata['aerosols'] = find(r'Aerosols considered: (.*)')
        metadata['aero_sources'] = find(r'Aerosols sources: (.*)')
        metadata['aero_abns'] = find(r'Aerosols abundances: (.*)')
        metadata['aero_abn_units'] = find(r'Aerosol abundance units: (.*)')
        metadata['aero_sizes'] = find(r'Aerosol sizes: (.*)')
        metadata['aero_size_units'] = find(r'Aerosol size units: (.*)')
        
        other_data['tab1_names'] = re.findall(r'#[ ]*(Alt\[km\].*)',text)[0]
        other_data['tab2_names'] = re.findall(r'#[ ]*(Alt\[km\].*)',text)[0]
        

        possible_table_lines =re.compile(r"(#[ ]+[\de\-\+\.\s]+)\n").findall(text,re.MULTILINE)
        tabs = []
        current_tab = []
        for line in possible_table_lines:
            if not '.' in line:
                if len(current_tab) == 0:
                    pass
                else:
                    tabs.append(current_tab)
                    current_tab = []
            else:
                current_tab.append(line)
        tab1_raw = '\n'.join(tabs[0])
        tab2_raw = '\n'.join(tabs[1])
                
        
        # tab1_raw = re.findall(
        #     r'Alt.*\n.*\n((# +\d[ \d.einf\-\+]+\n)+)',
        #     text
        # )[0]
        
        # tab2_raw = re.findall(
        #     r'Low.*\n.*\n((# +\d[ \d.einf\-\+]+\n)+)',
        #     text
        # )[0]
        
        integrated_vals = re.findall(
            r'Integrated[a-z ]+(.*)',text
        )[0]
        return metadata, other_data, tab1_raw, tab2_raw, integrated_vals
    
    @staticmethod
    def _get_tab_cols(
        raw_names:list,
        molec_unit : u.Unit,
        aero_unit : u.Unit,
        molecs:list = None,
        aeros:list = None
        )->Tuple[List[str],List[u.Unit]]:
        
        
        molecs = [] if molecs is None else molecs
        aeros = [] if aeros is None else aeros
        
        names = []
        units = []
        
        for i, name in enumerate(raw_names):
            try:
                unit = u.Unit(re.findall(r'\[(.*)\]',name)[0])
            except IndexError:
                unit = u.dimensionless_unscaled
            if name in molecs:
                names.append(name)
                units.append(molec_unit)
            elif name in aeros:
                names.append(name)
                units.append(aero_unit)
            elif 'size' in name:
                prev_name = raw_names[i-1]
                names.append(f'{prev_name}_size')
                units.append(unit)
            else:
                names.append(name.split('[')[0])
                units.append(unit)
        return names, units
    @staticmethod
    def _parse_tab_data(tab_raw:str)->np.ndarray:
        return np.array([
            np.fromstring(line[1:],sep=' ') for line in tab_raw.split('\n')
        ])
    
    @staticmethod
    def _build_tab(
        names:List[str],
        units:List[u.Unit],
        data:np.ndarray,
        meta:dict = None
    )->QTable:
        dat = {}
        for i,(name,unit) in enumerate(zip(names,units)):
            dat[name] = data[:,i]*unit
        return QTable(dat,meta=meta)
        
    
    
    @classmethod
    def from_bytes(cls, b: bytes):
        """
        From a bytes object read from a .lyr file
        """
        s = b.decode(settings.get_setting('encoding'))
        
        metadata, other_data, tab1_raw, tab2_raw, integrated_vals = cls._parse(s)
        tab1_names, tab1_units = cls._get_tab_cols(
            other_data['tab1_names'].split(),
            u.dimensionless_unscaled,
            u.dimensionless_unscaled,
            molecs=metadata['molecules'],
            aeros=metadata['aerosols']
        )
        tab1_dat = cls._parse_tab_data(tab1_raw)
        
        tab2_names, tab2_units = cls._get_tab_cols(
            other_data['tab2_names'].split(),
            u.m**-2,
            u.kg*u.m**-2,
            molecs=metadata['molecules'],
            aeros=metadata['aerosols']
        )
        tab2_dat = cls._parse_tab_data(tab2_raw)
        
        tab1 = cls._build_tab(
            tab1_names,
            tab1_units,
            tab1_dat,
            # meta=metadata
        )
        tab2 = cls._build_tab(
            tab2_names,
            tab2_units,
            tab2_dat,
        )
        return cls(
            prof=tab1,
            cg=tab2
        )
    def to_fits(self,path:Path):
        """
        Write to a fits file.
        """
        # hdulist = fits.HDUList()
        # prof_hdu = fits.table_to_hdu(self.prof)
        self.prof.write(path,overwrite=True)
        self.cg.write(path,append=True)
        # prof_hdu.header['EXTNAME'] = self.PROF_EXT
        # hdulist.append(prof_hdu)
        
        # cg_hdu = fits.table_to_hdu(self.cg)
        # cg_hdu.header['EXTNAME'] = self.CG_EXT
        # hdulist.append(cg_hdu)
        
        # hdulist.writeto(path,overwrite=True)
    @classmethod
    def from_fits(cls,path:Path):
        with fits.open(path) as hdulist:
            prof_table = QTable.read(hdulist[cls.PROF_EXT])
            cg_table = QTable.read(hdulist[cls.CG_EXT])
            
        return cls(prof=prof_table, cg=cg_table)
